package masterelection

import (
	"log"
	"time"

	"git.autistici.org/ale/autoradio/third_party/github.com/coreos/go-etcd/etcd"
)

const (
	STATE_SLAVE = iota
	STATE_MASTER
)

func stateToString(state int) string {
	switch state {
	case STATE_SLAVE:
		return "slave"
	case STATE_MASTER:
		return "master"
	}
	return ""
}

type MasterElection struct {
	client  *etcd.Client
	stop    chan bool
	stopped bool

	Addr string
	Path string
	TTL  uint64

	State       int
	StateChange chan int
}

func NewMasterElection(client *etcd.Client, path, addr string, ttl uint64, sch chan int, stop chan bool) *MasterElection {
	if ttl < 2 {
		ttl = 2
	}
	return &MasterElection{
		client:      client,
		Path:        path,
		Addr:        addr,
		TTL:         ttl,
		State:       STATE_SLAVE,
		StateChange: sch,
		stop:        stop,
	}
}

func (m *MasterElection) IsMaster() bool {
	return m.State == STATE_MASTER
}

func (m *MasterElection) GetMasterAddr() string {
	response, err := m.client.Get(m.Path, false, false)
	if err != nil || response.Node == nil {
		return ""
	}
	return response.Node.Value
}

func (m *MasterElection) setState(state int) {
	if m.State == state {
		return
	}
	log.Printf("masterelection: %s -> %s",
		stateToString(m.State),
		stateToString(state))
	if m.StateChange != nil {
		m.StateChange <- state
	}
	m.State = state
}

func (m *MasterElection) stopper() {
	<-m.stop

	// Tell the Run() goroutine to exit as soon as it can (we
	// could have simply used the 'stop' channel there but Watch()
	// ignores the stop channel if invoked without a receiving
	// channel and we'd have to refactor Run() to use a temp
	// channel for each invocation.
	m.stopped = true

	// Remove the lock file if we are the master.
	if m.State == STATE_MASTER {
		log.Printf("releasing masterelection lock")
		m.client.Delete(m.Path, false)
	}
}

func (m *MasterElection) runMaster(index uint64) {
	m.setState(STATE_MASTER)

	// If we renew the lease every TTL / N, we allow N renewal
	// errors before we stop believing being the master.
	ttl := time.Second * time.Duration(m.TTL)
	tick := time.NewTicker(ttl / 3)
	lastUpdate := time.Now()

	for {
		select {
		case t := <-tick.C:
			// To verify that we actually are still the
			// master (not just we believe we are), try
			// a compare-and-swap operation to check that
			// the stored master address is still our own,
			// and no-one stole our lock. If not, the TTL
			// will be updated (and the lock renewed).
			response, err := m.client.CompareAndSwap(m.Path, m.Addr, m.TTL, m.Addr, index)
			if err != nil {
				log.Printf("error updating lock: %s", err)

				// If we can't renew the lock for a
				// TTL, we must assume we lost it.
				if t.Sub(lastUpdate) > ttl {
					log.Printf("too many errors, lost lock")
					return
				}
			}
			index = response.EtcdIndex
			lastUpdate = t
		case <-m.stop:
			return
		}
	}
}

func (m *MasterElection) runSlave(index uint64) {
	m.setState(STATE_SLAVE)

	for {
		// Start a watch on the lock, waiting for its removal.
		response, err := m.client.Watch(m.Path, index+1, false, nil, m.stop)
		if err != nil {
			if err != etcd.ErrWatchStoppedByUser {
				log.Printf("slave Watch() error: %+v", err)
			}
			return
		}

		if response.Action == "delete" || response.Action == "expire" {
			return
		}

		index = response.EtcdIndex
	}

}

func (m *MasterElection) Run() {
	go m.stopper()

	// Start as a slave.
	m.setState(STATE_SLAVE)

	var watchIndex uint64

	for !m.stopped {

		// Since a failed Create does not return the
		// RAFT index, let's optimistically query the lock
		// before starting just to set a baseline for the
		// following Watch().
		if iresponse, err := m.client.Get(m.Path, false, false); err == nil {
			log.Printf("lock already exists: %+v", iresponse)
			watchIndex = iresponse.EtcdIndex
		}

		// Try to acquire the lock. This call will only succeed
		// if the lockfile does not exist (either because it
		// expired, or the previous master exited gracefully and
		// deleted it).
		response, err := m.client.Create(m.Path, m.Addr, m.TTL)

		if err == nil {
			// Howdy, we're the master now. Wait a while
			// and renew our TTL.
			log.Printf("masterelection: we are the master")
			m.runMaster(response.EtcdIndex)
		} else {
			// We're not the master. Wait until the lock
			// is deleted or expires.
			log.Printf("masterelection: running as slave (%s)", err)
			m.runSlave(watchIndex)
		}
	}
}
